from typing									import Any
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import patronus








class ARPSniffer(Transmutable):

	"""
		Link layer class for listening broadcast ARP traffic. Must accept callable object "sniffer", which
		once provided with all keyword arguments "kwargs" must establish an ARP trap. Implementation of
		"sniffer" must account for all needs, such as working time or received packets handlers, which
		might be provided with "kwargs". Any Exception raised during "sniffer" working will be caught
		and logged, and this will cause "sniffer" to stop.
	"""

	def __call__(self, sniffer :Callable[[Any],Any], **kwargs):


		if	callable(sniffer):


			self.loggy.info(f"Establishing ARP trap")
			for k,v in kwargs.items() : self.loggy.debug(f"Using {k}: {v}")


			try:	sniffer(**kwargs)
			except	Exception as E : self.loggy.debug(f"ARP trap failed due to {patronus(E)}")
			else:	self.loggy.info("ARP trap ended")
		else:		self.loggy.debug(f"Invalid sniffer \"{sniffer}\"")










import	os
import	unittest
from	pygwarts.tests.filch		import FilchTestCase
from	pygwarts.irma.contrib		import LibraryContrib
from	pygwarts.filch.apppuf.snmp	import SNMPtrap








class ApppufCase(FilchTestCase):

	"""
		Testing L5 instruments
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.APPPUF_HANDLER): os.remove(cls.APPPUF_HANDLER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.APPPUF_HANDLER)
	def test_SNMPtrap(self):

		class TestListener(SNMPtrap):
			class loggy(LibraryContrib):

				handler		= self.APPPUF_HANDLER
				init_name	= "SNMPtrap"
				init_level	= 10

		self.test_case = TestListener()

		with self.assertLogs("SNMPtrap", 10) as case_loggy:
			self.test_case(

				"127.0.0.11", 54321,
				lambda *args, **kwargs : None,
				listen_timer=42,
			)

		self.assertIn("INFO:SNMPtrap:Establishing SNMP trap at 127.0.0.11:54321", case_loggy.output)
		self.assertIn("DEBUG:SNMPtrap:Using listen_timer: 42", case_loggy.output)
		self.assertIn("INFO:SNMPtrap:127.0.0.11:54321 SNMP trap discarded", case_loggy.output)




	def test_SNMPtrap_invalid_ip(self):

		class TestListener(SNMPtrap):
			class loggy(LibraryContrib):

				handler		= self.APPPUF_HANDLER
				init_name	= "SNMPtrap_invalid_ip"
				init_level	= 10

		self.test_case = TestListener()

		with self.assertLogs("SNMPtrap_invalid_ip", 10) as case_loggy:
			self.test_case("127.0.0.311", 54321, lambda *args : None)

		self.assertIn("DEBUG:SNMPtrap_invalid_ip:Listen ip verification failed", case_loggy.output)




	def test_SNMPtrap_invalid_port(self):

		class TestListener(SNMPtrap):
			class loggy(LibraryContrib):

				handler		= self.APPPUF_HANDLER
				init_name	= "SNMPtrap_invalid_port"
				init_level	= 10

		self.test_case = TestListener()

		with self.assertLogs("SNMPtrap_invalid_port", 10) as case_loggy:
			self.test_case("127.0.0.11", 543211, lambda *args : None)

		self.assertIn("DEBUG:SNMPtrap_invalid_port:Listen port verification failed", case_loggy.output)




	def test_SNMPtrap_raise(self):

		class TestListener(SNMPtrap):
			class loggy(LibraryContrib):

				handler		= self.APPPUF_HANDLER
				init_name	= "SNMPtrap_raise"
				init_level	= 10

		def fake_trap(*args): raise ValueError("Enough of this")
		self.test_case = TestListener()

		with self.assertLogs("SNMPtrap_raise", 10) as case_loggy:
			self.test_case("127.0.0.11", 54321, fake_trap)

		self.assertIn(
			f"ERROR:SNMPtrap_raise:{self.test_case} got ValueError: Enough of this", case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib									import Path
from	typing									import Dict
from	pygwarts.tests.hagrid					import MediumSet
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.irma.shelve					import LibraryShelf
from	pygwarts.hagrid.thrivables				import Tree
from	pygwarts.hagrid.sprouts					import fssprout
from	pygwarts.hagrid.planting				import Flourish
from	pygwarts.hagrid.cultivation.registering	import PlantRegister
from	pygwarts.hagrid.cultivation.registering	import TG
from	pygwarts.hagrid.cultivation.registering	import LG
from	pygwarts.hagrid.cultivation.registering	import WG
from	pygwarts.hagrid.cultivation.registering	import TL
from	pygwarts.hagrid.cultivation.registering	import LL
from	pygwarts.hagrid.cultivation.registering	import WL








class RegisteringCases(MediumSet):

	"""
		Complex testing of cultivation.registering functionality
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.MEDIUM_HANDLER_8): os.remove(cls.MEDIUM_HANDLER_8)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.MEDIUM_HANDLER_8)

	def measures_fields_types_check(self, folder :Dict[str,int]):

		self.assertIsInstance(folder.get(TG), int)
		self.assertIsInstance(folder.get(LG), int)
		self.assertIsInstance(folder.get(WG), int)
		self.assertIsInstance(folder.get(TL), int)
		self.assertIsInstance(folder.get(LL), int)
		self.assertIsInstance(folder.get(WL), int)

	def all_sources_exists(self):

		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))

	def first_sprout_real(self):

		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_1), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_1)), 2)
		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_1).get(self.d_mss1_md), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md]), 4)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md].get(self.f_mss1_md_rock), int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md].get(self.f_mss1_md_paper), int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md].get(self.f_mss1_md_scissors), int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md].get(self.d_mss1_md_gc), dict
		)
		self.assertEqual(
			len(self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_md][self.d_mss1_md_gc]), 3
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_1
			)[	self.d_mss1_md
			][	self.d_mss1_md_gc].get(self.f_mss1_md_gc_sadmovies), int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_1
			)[	self.d_mss1_md
			][	self.d_mss1_md_gc].get(self.f_mss1_md_gc_dumbasses), int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_1
			)[	self.d_mss1_md
			][	self.d_mss1_md_gc].get(self.f_mss1_md_gc_onion), int
		)
		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_1).get(self.d_mss1_c), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c]), 8)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_good), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_good
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_notgood), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_notgood
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_bad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_bad
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_notbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_notbad
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_notnot), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_notnot
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_badbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_badbad
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_badgood), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_badgood
				]
			),		0
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_1)[self.d_mss1_c].get(self.d_mss1_c_goodbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(

					self.MEDIUM_SET_SPROUT_1
				)[	self.d_mss1_c
				][	self.d_mss1_c_goodbad
				]
			),		0
		)

	def second_sprout_real(self):

		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_2), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_2)), 3)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_2).get(self.f_mss2_tasksschedule), int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_2).get(self.f_mss2_explangraph), int
		)
		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_2).get(self.d_mss2_c), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c]), 8)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_hardcleaning),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_hardwashing),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_mediumcooking),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_mediumhomework),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_mediumphysicals),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_2
			)[	self.d_mss2_c
			].get(
				self.f_mss2_c_easyprocrastinating
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(self.MEDIUM_SET_SPROUT_2)[self.d_mss2_c].get(self.f_mss2_c_easyreflections),
			int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_2
			)[	self.d_mss2_c
			].get(
				self.f_mss2_c_easyphilosophizing
			),	int
		)

	def third_sprout_real(self):

		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_3), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_3)), 1)
		self.assertIsInstance(self.test_case.garden(self.MEDIUM_SET_SPROUT_3).get(self.d_mss3_k), dict)
		self.assertEqual(len(self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k]), 3)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k].get(self.d_mss3_k_f), dict
		)
		self.assertEqual(
			len(self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k][self.d_mss3_k_f]), 3
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_f
			].get(
				self.f_mss3_k_f_milk
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_f
			].get(
				self.f_mss3_k_f_cheese
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_f
			].get(
				self.f_mss3_k_f_meat
			),	int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k].get(self.d_mss3_k_o), dict
		)
		self.assertEqual(
			len(self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k][self.d_mss3_k_o]), 2
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_o
			].get(
				self.f_mss3_k_o_chicken
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_o
			].get(
				self.f_mss3_k_o_pie
			),	int
		)
		self.assertIsInstance(
			self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k].get(self.d_mss3_k_t), dict
		)
		self.assertEqual(
			len(self.test_case.garden(self.MEDIUM_SET_SPROUT_3)[self.d_mss3_k][self.d_mss3_k_t]), 1
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			].get(
				self.d_mss3_k_t_bc
			),	dict
		)
		self.assertEqual(
			len(
				self.test_case.garden(
					self.MEDIUM_SET_SPROUT_3
				)[	self.d_mss3_k
				][	self.d_mss3_k_t
				][	self.d_mss3_k_t_bc
				]
			),	5
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			][	self.d_mss3_k_t_bc
			].get(
				self.f_mss3_k_t_bc_bread
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			][	self.d_mss3_k_t_bc
			].get(
				self.f_mss3_k_t_bc_crumb1
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			][	self.d_mss3_k_t_bc
			].get(
				self.f_mss3_k_t_bc_crumb2
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			][	self.d_mss3_k_t_bc
			].get(
				self.f_mss3_k_t_bc_crumb420
			),	int
		)
		self.assertIsInstance(

			self.test_case.garden(

				self.MEDIUM_SET_SPROUT_3
			)[	self.d_mss3_k
			][	self.d_mss3_k_t
			][	self.d_mss3_k_t_bc
			].get(
				self.f_mss3_k_t_bc_crumb69
			),	int
		)

	def first_sprout_magical(self):

		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_1], dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_1]), 8)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_1])
		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_1].get(self.d_mss1_md), dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_md]), 7)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_md])
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_md].get(self.d_mss1_md_gc), dict
		)
		self.assertEqual(
			len(self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_md][self.d_mss1_md_gc]), 6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_md][self.d_mss1_md_gc]
		)
		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_1].get(self.d_mss1_c), dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c]), 14)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c])
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_good), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_good]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_good]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_notgood), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_notgood]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_notgood]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_bad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_bad]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_bad]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_notbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_notbad]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_notbad]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_notnot), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_notnot]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_notnot]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_badbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_badbad]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_badbad]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_badgood), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_badgood]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_badgood]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c].get(self.d_mss1_c_goodbad), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_1][
				self.d_mss1_c][
				self.d_mss1_c_goodbad]
			),	6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_1][self.d_mss1_c][self.d_mss1_c_goodbad]
		)


	def second_sprout_magical(self):

		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_2], dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_2]), 7)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_2])
		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_2].get(self.d_mss2_c), dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_2][self.d_mss2_c]), 6)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_2][self.d_mss2_c])


	def third_sprout_magical(self):

		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_3], dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_3]), 7)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_3])
		self.assertIsInstance(self.test_case.garden[self.MEDIUM_SET_SPROUT_3].get(self.d_mss3_k), dict)
		self.assertEqual(len(self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k]), 9)
		self.measures_fields_types_check(self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k])
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k].get(self.d_mss3_k_f), dict
		)
		self.assertEqual(
			len(self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_f]), 6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_f]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k].get(self.d_mss3_k_o), dict
		)
		self.assertEqual(
			len(self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_o]), 6
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_o]
		)
		self.assertIsInstance(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k].get(self.d_mss3_k_t), dict
		)
		self.assertEqual(
			len(self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_t]), 7
		)
		self.measures_fields_types_check(
			self.test_case.garden[self.MEDIUM_SET_SPROUT_3][self.d_mss3_k][self.d_mss3_k_t]
		)
		self.assertIsInstance(

			self.test_case.garden[
			self.MEDIUM_SET_SPROUT_3][
			self.d_mss3_k][
			self.d_mss3_k_t].get(self.d_mss3_k_t_bc), dict
		)
		self.assertEqual(
			len(
				self.test_case.garden[
				self.MEDIUM_SET_SPROUT_3][
				self.d_mss3_k][
				self.d_mss3_k_t][
				self.d_mss3_k_t_bc]
			),	6
		)
		self.measures_fields_types_check(

			self.test_case.garden[
			self.MEDIUM_SET_SPROUT_3][
			self.d_mss3_k][
			self.d_mss3_k_t][
			self.d_mss3_k_t_bc]
		)




	def test_bottom_register(self):
		class Sakura(Tree):

			class loggy(LibraryContrib):

				init_name	= "bottom_register"
				init_level	= 10
				handler		= self.MEDIUM_HANDLER_8

			@fssprout(self.MEDIUM_SET_SPROUT_1)
			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@fssprout(self.MEDIUM_SET_SPROUT_3)
			@PlantRegister("garden")
			class sync(Flourish):		pass
			class garden(LibraryShelf):	pass


		self.test_case = Sakura()
		self.test_case.sync()


		self.all_sources_exists()
		self.assertEqual(len(self.test_case.garden), 3)
		self.assertEqual(len(self.test_case.garden()), 3)


		self.first_sprout_real()
		self.second_sprout_real()
		self.third_sprout_real()
		self.first_sprout_magical()
		self.second_sprout_magical()
		self.third_sprout_magical()




	def test_middle_register(self):
		class Sakura(Tree):

			class loggy(LibraryContrib):

				init_name	= "middle_register"
				init_level	= 10
				handler		= self.MEDIUM_HANDLER_8

			@fssprout(self.MEDIUM_SET_SPROUT_1)
			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@PlantRegister("garden")
			@fssprout(self.MEDIUM_SET_SPROUT_3)
			class sync(Flourish):		pass
			class garden(LibraryShelf):	pass


		self.test_case = Sakura()
		self.test_case.sync()


		self.all_sources_exists()
		self.assertEqual(len(self.test_case.garden), 2)
		self.assertEqual(len(self.test_case.garden()), 2)


		self.first_sprout_real()
		self.second_sprout_real()
		self.first_sprout_magical()
		self.second_sprout_magical()


		self.assertIsNone(self.test_case.garden[self.MEDIUM_SET_SPROUT_3])
		self.assertIsNone(self.test_case.garden()[self.MEDIUM_SET_SPROUT_3])




	def test_upper_register(self):
		class Sakura(Tree):

			class loggy(LibraryContrib):

				init_name	= "upper_register"
				init_level	= 10
				handler		= self.MEDIUM_HANDLER_8

			@PlantRegister("garden")
			@fssprout(self.MEDIUM_SET_SPROUT_1)
			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@fssprout(self.MEDIUM_SET_SPROUT_3)
			class sync(Flourish):		pass
			class garden(LibraryShelf):	pass


		self.test_case = Sakura()
		self.test_case.sync()


		self.all_sources_exists()
		self.assertEqual(len(self.test_case.garden), 0)
		self.assertEqual(len(self.test_case.garden()), 0)


		self.assertIsNone(self.test_case.garden[self.MEDIUM_SET_SPROUT_1])
		self.assertIsNone(self.test_case.garden[self.MEDIUM_SET_SPROUT_2])
		self.assertIsNone(self.test_case.garden[self.MEDIUM_SET_SPROUT_3])
		self.assertIsNone(self.test_case.garden()[self.MEDIUM_SET_SPROUT_1])
		self.assertIsNone(self.test_case.garden()[self.MEDIUM_SET_SPROUT_2])
		self.assertIsNone(self.test_case.garden()[self.MEDIUM_SET_SPROUT_3])




	def test_free_register(self):
		class Sakura(Tree):

			class loggy(LibraryContrib):

				init_name	= "free_register"
				init_level	= 10
				handler		= self.MEDIUM_HANDLER_8

			@PlantRegister("garden")
			class sync(Flourish):		pass
			class garden(LibraryShelf):	pass


		self.test_case = Sakura()
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_1,
			Path(self.MEDIUM_SET_SPROUT_1),
			[ Path(self.d_mss1_md), Path(self.d_mss1_c) ],
			[]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_1,
			Path(self.d_mss1_md),
			[ Path(self.d_mss1_md_gc) ],
			[
				Path(self.f_mss1_md_rock),
				Path(self.f_mss1_md_paper),
				Path(self.f_mss1_md_scissors),
			]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_1,
			Path(self.d_mss1_md_gc),
			[],
			[
				Path(self.f_mss1_md_gc_sadmovies),
				Path(self.f_mss1_md_gc_dumbasses),
				Path(self.f_mss1_md_gc_onion),
			]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_1,
			Path(self.d_mss1_c),
			[
				Path(self.d_mss1_c_good),
				Path(self.d_mss1_c_notgood),
				Path(self.d_mss1_c_bad),
				Path(self.d_mss1_c_notbad),
				Path(self.d_mss1_c_notnot),
				Path(self.d_mss1_c_badbad),
				Path(self.d_mss1_c_badgood),
				Path(self.d_mss1_c_goodbad),
			],
			[]
		)
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_good), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_notgood), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_bad), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_notbad), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_notnot), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_badbad), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_badgood), [], [])
		self.test_case.sync(self.MEDIUM_SET_SPROUT_1, Path(self.d_mss1_c_goodbad), [], [])


		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_2,
			Path(self.MEDIUM_SET_SPROUT_2),
			[ Path(self.d_mss2_c) ],
			[ Path(self.f_mss2_tasksschedule), Path(self.f_mss2_explangraph) ]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_2,
			Path(self.d_mss2_c),
			[],
			[
				Path(self.f_mss2_c_hardcleaning),
				Path(self.f_mss2_c_hardwashing),
				Path(self.f_mss2_c_mediumcooking),
				Path(self.f_mss2_c_mediumhomework),
				Path(self.f_mss2_c_mediumphysicals),
				Path(self.f_mss2_c_easyprocrastinating),
				Path(self.f_mss2_c_easyreflections),
				Path(self.f_mss2_c_easyphilosophizing)
			]
		)


		self.test_case.sync(
			self.MEDIUM_SET_SPROUT_3, Path(self.MEDIUM_SET_SPROUT_3), [ Path(self.d_mss3_k) ], []
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_3,
			Path(self.d_mss3_k),
			[ Path(self.d_mss3_k_f), Path(self.d_mss3_k_o), Path(self.d_mss3_k_t) ],
			[]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_3,
			Path(self.d_mss3_k_f),
			[],
			[ Path(self.f_mss3_k_f_milk), Path(self.f_mss3_k_f_cheese), Path(self.f_mss3_k_f_meat) ]
		)
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_3,
			Path(self.d_mss3_k_o),
			[],
			[ Path(self.f_mss3_k_o_chicken), Path(self.f_mss3_k_o_pie) ]
		)
		self.test_case.sync(self.MEDIUM_SET_SPROUT_3, Path(self.d_mss3_k_t), [ Path(self.d_mss3_k_t_bc) ], [])
		self.test_case.sync(

			self.MEDIUM_SET_SPROUT_3,
			Path(self.d_mss3_k_t_bc),
			[],
			[
				Path(self.f_mss3_k_t_bc_bread),
				Path(self.f_mss3_k_t_bc_crumb1),
				Path(self.f_mss3_k_t_bc_crumb2),
				Path(self.f_mss3_k_t_bc_crumb420),
				Path(self.f_mss3_k_t_bc_crumb69)
			]
		)


		self.first_sprout_real()
		self.second_sprout_real()
		self.third_sprout_real()
		self.first_sprout_magical()
		self.second_sprout_magical()
		self.third_sprout_magical()








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing												import Any
from typing												import Optional
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.magical.spells							import flagrate
from pygwarts.irma.shelve								import LibraryShelf
from pygwarts.irma.access.volume						import LibraryVolume
from pygwarts.irma.shelve.casing						import shelf_case








class ViewWrapper(ControlledTransmutation):

	"""
		irma.access utility decorator, that serves as an instrument for text editing, which is just addition
		of a text before (header) and after (footer), specially for VolumeBookmark objects. As
		ControlledTransmutation class, accepts corresponding "header" and "footer" arguments, which must be
		strings and defaulted to empty strings, in case only one is needed. In mutable chain acts as a
		mutation - takes decorated Transmutable, which must be a VolumeBookmark class with "view" method
		defined, and overwrites it to return wrapped string. Returns final string, if super.view returns
		string, or None otherwise, as super.view must do. If "view" not found as decorated Transmutable
		class member, no mutation will happen and original Transmutable will be propagated. Doesn't handles
		any Exceptions, only explicitly converts header" and "footer" to strings at initiation time and
		once again in decoration's interpolation.
	"""

	def __init__(self, header :str =str(), footer :str =str()):

		self.header = str(header)
		self.footer = str(footer)

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		header = self.header
		footer = self.footer


		if	hasattr(layer, "view"):
			class ViewWrapping(geminio(layer)):

				def view(self, volume :LibraryVolume) -> str | None :
					if	isinstance(current_view := super().view(volume), str):

						final_value = f"{header}{current_view}{footer}"
						vlen = len(final_value)
						self.loggy.debug(f"Wrapped view now {vlen} symbol{flagrate(vlen)}")


						return	final_value


					self.loggy.debug("View to wrap not found")
			return	ViewWrapping
		return		layer








class ViewCase(ControlledTransmutation):

	"""
		irma.shelve utility decorator, that integrates the core "casing" functionality into mutable chain,
		by maintaining VolumeBookmark "view" method return. As ControlledTransmutation class, accepts
		positional argument "link", which must be a string that represents the name of LibraryShelf
		attribute, and optional keyword only arguments "prep" and "post", which must be "casing" handling
		callables. The key for LibraryShelf will be a string representation of current mutable chain point,
		and it will be passed along with this three accepted arguments to "shelf_case" function. In
		mutable chain acts as a mutation - takes decorated Transmutable, which must be a VolumeBookmark
		class with "view" method defined, and overwrites it to pass it's returned string value to
		"shelf_case" function. If "shelf_case" return value is not None it will be returned. Otherwise
		returns original "view" string. Returns None if overwritten "view" returns None. Despite "view"
		always must return string, this decoration doesn't restrict "post" return value, but suggests
		it will either return a string or will be passed to another object, which will return string.
	"""

	def __init__(
					self,
					link	:str,
					*,
					prep	:Optional[Callable[[str],str]]	=(lambda T	: T),
					post	:Optional[Callable[[str],str]]	=(lambda T,P: T),
				):

		self.link = link
		self.prep = prep
		self.post = post

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		link = self.link
		post = self.post
		prep = self.prep


		if	hasattr(layer, "view"):
			class ViewCasing(geminio(layer)):

				def view(self, volume :LibraryVolume) -> Any | None :
					if	isinstance(current_view := super().view(volume), str):
						if	isinstance(mapping := getattr(self, str(link), None), LibraryShelf | dict):


							if	str(volume) not in mapping : mapping[str(volume)] = dict()
							if	(casing := shelf_case(

								current_view,
								key=str(self),
								shelf=mapping[str(volume)],
								prep=prep,
								post=post,

							))	is not None:

								self.loggy.debug(f"Obtained view casing value \"{casing}\"")
								return casing


						self.loggy.debug("View casing value not obtained")
						return current_view


					self.loggy.debug("View to case not found")
			return	ViewCasing
		return		layer







